/*
  This script creates stored procedure sp_purge_audit for purging audit records.

  sp_purge_audit
  Purge audit records when designrevisionentity.processStatus is not null.
  The following tables are affected by the purging script: 
    auditingentry 
    revision_auditingentry 
    All table with _aud suffix 
        The following tables are excluded: 
          health_record_aud 
          health_round_samples_aud 
          health_sample_aud 
          health_threshold_aud
          published_flow_aud

  Parameters:
  time_limit         Number of seconds after which no command is executed.  The procedure
                     will stop when the time limit is reached and the last command
                     (started before the time limit) is complete.  The default is no time 
                     limit (null).
  time_delay         Number of seconds to delay between each delete command.  The default
                     is null, no delay.
  delete_chunk_size  Maximum number of rows to delete per command.  The default is 10000
                     rows.

  Example:
  The following example purges audit records with a time limit so that no command is
  executed after 3600 seconds (1 hour)

  exec sp_purge_audit @time_limit=3600
*/

IF OBJECT_ID ( 'dbo.sp_purge_audit', 'P' ) IS NOT NULL 
    DROP PROCEDURE dbo.sp_purge_audit;
GO

CREATE PROCEDURE dbo.sp_purge_audit
  @time_limit int = null,
  @time_delay int = null,
  @delete_chunk_size int = 10000
AS
DECLARE @aud_table TABLE(parent_table sysname, 
                         child_table sysname,
                         level int,
                         primary key(parent_table, child_table));

DECLARE cur_aud_table CURSOR FOR
  SELECT parent_table, child_table FROM @aud_table order by parent_table, level desc, child_table;

DECLARE @new_line nvarchar(2) = char(13) + char(10);
DECLARE @stop_time datetime;
DECLARE @sql nvarchar(max);
DECLARE @parent_table sysname;
DECLARE @child_table sysname;
DECLARE @prev_parent_table sysname;
DECLARE @delay_stmt nvarchar(300) = N'';

BEGIN
  SET NOCOUNT ON;
  SET IMPLICIT_TRANSACTIONS OFF;
  SET XACT_ABORT ON;

  set @stop_time = case when @time_limit is null then null else dateadd(second, @time_limit, getdate()) end;

  print convert(varchar(30), getdate(), 109) + ': Deleting audit entries...';

  if @time_delay is not null
    set @delay_stmt = @new_line
                      + N'  WAITFOR DELAY ''' 
                        + RIGHT(N'0' + cast(FLOOR(@time_delay/3600.0) as nvarchar), 2) + ':'
                        + RIGHT(N'0' + cast(FLOOR((@time_delay % 3600)/60.0) as nvarchar), 2) + ':'
                        + RIGHT(N'0' + cast(@time_delay % 60 as nvarchar), 2) + '''';

  -- populate @aud_tables with _aud table names based on FK relationship
  -- The following tables should not be purged and are excluded from the query:
  --   published_flow_aud, health_record_aud, health_round_samples_aud, health_threshold_aud, health_sample_aud
  with t1 as
  ( select  t.object_id foreign_table_object_id
            , t.object_id primary_table_object_id
            , 0 level
            , cast(null as sysname) foreign_key_name
    from sys.tables t 
    where t.name like '%[_]aud'  /* parent table */
    and not exists (select 1
               from sys.foreign_keys fk, sys.tables pt
               where t.object_id = fk.parent_object_id
               and pt.object_id = fk.referenced_object_id 
               and pt.name like '%[_]aud')
    UNION ALL
    select fk.parent_object_id foreign_table_object_id
            , t1.primary_table_object_id
            , t1.level + 1 level
            , fk.name foreign_key_name
    from t1, sys.foreign_keys fk
    where fk.referenced_object_id = t1.foreign_table_object_id
  )
  insert into @aud_table (parent_table, child_table, level)
  select pt.name parent_table, t.name child_table, level 
  from t1, sys.tables t
       join sys.columns c
         on t.object_id = c.object_id and lower(c.name) = 'rev'
       , sys.tables pt
  where t.object_id = t1.foreign_table_object_id
  and t.name not in ('published_flow_aud', 'health_record_aud', 'health_round_samples_aud', 'health_threshold_aud', 'health_sample_aud')
  and pt.object_id = primary_table_object_id;


  -- create temp tables to store ID and REV to be deleted
  CREATE TABLE #aud_id_rev__
  (id numeric(19,0),
   rev int
   primary key(id, rev));

  -- loop to delete from parent and child aud tables until time limit is reached
  OPEN cur_aud_table;
  FETCH NEXT FROM cur_aud_table INTO @parent_table, @child_table;

  set @prev_parent_table = N' ';

  WHILE @@FETCH_STATUS = 0 AND (@stop_time is null OR getdate() < @stop_time) BEGIN

    IF @prev_parent_table <> @parent_table and @parent_table = @child_table BEGIN
      -- table has no child.  Loop and delete a chunk at a time until time limit is reached
      set @sql = N'WHILE (@stop_time is null OR getdate() < @stop_time) BEGIN'
                 + @delay_stmt
                 + @new_line
                 + N'  delete top(' + cast(@delete_chunk_size as nvarchar(10)) + N') aud'
                 + N'  from ' + @parent_table + N' aud, designrevisionentity d'
                 + N'  where aud.rev = d.id'
                 + N'  and d.processStatus is not null'
                 + N'  OPTION (FORCE ORDER, LOOP JOIN);'
                 + @new_line
                 + N'  IF @@ROWCOUNT < ' + cast(@delete_chunk_size as nvarchar(10)) + ' BREAK;'
                 + @new_line
                 + N'END';

      PRINT @sql;

      EXEC sp_executesql @sql,
            N'@stop_time DATETIME',
            @stop_time;
    END
    ELSE BEGIN 
      -- delete from process aud table and its children
      IF @prev_parent_table <> @parent_table BEGIN

        set @sql = N'WHILE (@stop_time is null OR getdate() < @stop_time) BEGIN'
                   + @new_line
                   + N'  truncate table #aud_id_rev__;'
                   + @new_line
                   + N'  INSERT INTO #aud_id_rev__ select top(' + cast(@delete_chunk_size as nvarchar(10)) + ') aud.ID, aud.REV'
                   + N' from ' + @parent_table + ' aud, designrevisionentity d'
                   + N' where aud.rev = d.id'
                   + N' and d.processStatus is not null'
                   + N' OPTION (FORCE ORDER, LOOP JOIN);'
                   + @new_line
                   + N'  IF @@ROWCOUNT = 0 BREAK;';
      END

      set @sql = @sql 
                 + @delay_stmt
                 + @new_line
                 + N'  delete aud'
                 + ' from #aud_id_rev__ t, ' + @child_table + ' aud'
                 + ' where aud.id = t.id and aud.rev = t.rev'
                 + ' OPTION (FORCE ORDER, LOOP JOIN);';

      IF @parent_table = @child_table BEGIN
        set @sql = @sql + @new_line
                   + N'END';

        PRINT @sql;

        EXEC sp_executesql @sql,
              N'@stop_time DATETIME',
              @stop_time;
      END;
    END;

    set @prev_parent_table = @parent_table;
    
    FETCH NEXT FROM cur_aud_table INTO @parent_table, @child_table;
  END;

  CLOSE cur_aud_table;
  DEALLOCATE cur_aud_table;

  DROP TABLE #aud_id_rev__;


  IF (@stop_time is null OR getdate() < @stop_time) BEGIN

      -- create temp tables to store audit record ID's to be deleted
      CREATE TABLE #revision_auditingentry__
      (designrevisionentity_id int, 
       auditEntries_id numeric(19,0),
       primary key(auditEntries_id));

      -- loop to delete rows of audit entries until time limit is reached
      -- the loop will break if no record is found
      set @sql =   N'WHILE (@stop_time is null OR getdate() < @stop_time) BEGIN'
                 + @new_line
                 + N'  truncate table #revision_auditingentry__;'
                 + @new_line
             
                 -- get ID's from audit entry tables and store in #revision_auditingentry__
                 + N'  INSERT into #revision_auditingentry__'
                 + N' select top(' + cast(@delete_chunk_size as nvarchar(10)) 
                 + N') r.designrevisionentity_id, r.auditEntries_id'
                 + N' from revision_auditingentry r, designrevisionentity d'
                 + N' where r.designrevisionentity_id = d.id'
                 + N' and d.processStatus is not null'
                 + N' OPTION (FORCE ORDER, LOOP JOIN);'
                 + @new_line

                 -- break if no row found
                 + N'  if @@ROWCOUNT = 0'
                 + N' break;'

                 -- delete revision_auditingentry
                 + @delay_stmt
                 + @new_line
                 + N'  delete revision_auditingentry'
                 + N' from #revision_auditingentry__ r, revision_auditingentry'
                 + N' where revision_auditingentry.designrevisionentity_id = r.designrevisionentity_id'
                 + N' and revision_auditingentry.auditEntries_id = r.auditEntries_id'
                 + N' OPTION (FORCE ORDER, LOOP JOIN);'

                 -- delete auditingentry
                 + @delay_stmt
                 + @new_line
                 + N'  delete from auditingentry where id in (select auditEntries_id from #revision_auditingentry__);'
                 
                 + N'END;';

      PRINT @sql;

      EXEC sp_executesql @sql,
            N'@stop_time DATETIME',
            @stop_time;

      DROP TABLE #revision_auditingentry__;
  END;

  if (getdate() > @stop_time)
    print 'Exceeded Time limit.  Exiting procedure.';
  else
    print 'Purge done.';
END;
GO



-------------------------------------
-- create index on REV column of each _aud table if it does not exist
-------------------------------------
DECLARE @sql nvarchar(1000)
DECLARE @table_name sysname
DECLARE @column_name1 sysname
DECLARE @column_name2 sysname
DECLARE @index_name sysname
DECLARE @tab_missing_index table(table_name sysname, column_name1 sysname, column_name2 sysname)

INSERT INTO @tab_missing_index
select t.name, c.name, c2.name
from sys.tables t 
      inner join sys.foreign_keys f on f.parent_object_id = t.object_id
	  inner join sys.tables t_parent on f.referenced_object_id = t_parent.object_id
      inner join sys.foreign_key_columns fc on fc.constraint_object_id = f.object_id and fc.constraint_column_id = 1
      inner join sys.columns c on c.object_id = t.object_id and c.column_id = fc.parent_column_id
	  left join sys.index_columns ic on ic.object_id = c.object_id and ic.column_id = c.column_id and ic.index_column_id = fc.constraint_column_id
      left join sys.foreign_key_columns fc2 on fc2.constraint_object_id = f.object_id and fc2.constraint_column_id = 2
      left join sys.columns c2 on c2.object_id = t.object_id and c2.column_id = fc2.parent_column_id
	  left join sys.index_columns ic2 on ic2.object_id = c2.object_id and ic2.index_id = ic.index_id and ic2.column_id = c2.column_id and ic2.index_column_id = fc2.constraint_column_id
where (ic.column_id is null or (c2.object_id is not null and  ic2.column_id is null))
and t.name like '%[_]aud'
and lower(c.name) = 'id' and lower(c2.name) = 'rev'
order by 1

DECLARE cur1 CURSOR FOR
SELECT table_name, column_name1, @column_name2 FROM @tab_missing_index
OPEN cur1
FETCH NEXT FROM cur1 INTO @table_name, @column_name1, @column_name2
WHILE @@FETCH_STATUS = 0
BEGIN
  set @index_name = 'IX_'+ @table_name + '_' + @column_name1 + '_' + @column_name2;

    -- check if there is already an index with the same name
  if exists(select 1 from sys.indexes where name = @index_name) 
  BEGIN
    -- index name is already in use.  Generate another name with a numeric suffix.
    -- e.g. IX_tab_REV_1, IX_tab_REV_1, etc. 
    select @index_name = @index_name + '_' + 
                           cast(case when num1 < 9 then num1
                                  when num2 < 99 then num2
                                  else num3 end + 1 as nvarchar(10))
      from
        (select ISNULL(MAX(cast(right(name, 1) as int)),0) num1
        FROM sys.indexes i
        WHERE i.name between @index_name + '_1' and @index_name+'_9') t1,
        (select ISNULL(MAX(cast(right(name, 2) as int)),9) num2
        FROM sys.indexes i
        WHERE i.name between @index_name + '_10' and @index_name + '_99') t2,
        (select ISNULL(MAX(cast(right(name, 3) as int)),99) num3
        FROM sys.indexes i
        WHERE i.name between @index_name + '_100' and @index_name + '_999') t3;
  END;

  set @sql = 'CREATE INDEX ' + @index_name + ' ON ' + @table_name + '(' + @column_name1 + ',' + @column_name2 + ')';
  PRINT @sql
  EXEC sp_executesql @sql

  FETCH NEXT FROM cur1 INTO @table_name, @column_name1, @column_name2;
END

CLOSE cur1
DEALLOCATE cur1

GO
