/*
  This script creates stored procedure sp_purge_execution_jobs for purging offline execution jobs.

  sp_purge_execution_jobs
  Purge execution jobs and its associated records one day at a time starting from the oldest
  records.  The following tables are being purged:
    offline_distribution_events
    offline_exec_param_links
    offline_exec_servers
    offline_execution_jobs
    offline_flow_events
    offline_manual_operations
    offline_parameter_requests
    offline_parameters
    offline_propagation_events
    offline_step_events
    offline_steps
    offline_user_events

  Parameters:
  retention_days       Number of days to retain execution job records.  Default is 90 days
  time_limit           Number of seconds after which no command is executed.  The procedure
                       will stop when the time limit is reached and the last command
                       (started before the time limit) is complete.  The default is no time 
                       limit (null).
  time_delay           Number of seconds to delay between each delete command.  The default
                       is null, no delay.
  delete_chunk_size    Maximum number of rows to delete per command.  The default is 10000
                       rows.

  Example:
  The following example purges jobs older than 120 days, with a time limit so that no command is
  executed after 3600 seconds (1 hour)

  exec sp_purge_execution_jobs @retention_days=120, @time_limit=3600

*/

-- create an index on offline_execution_jobs.end_time and id if it doesn't exists
if not exists 
     (select 1 
      from sys.tables t join sys.columns c1 
                on t.object_id = c1.object_id
                and UPPER(c1.name) = 'END_TIME'
            join sys.index_columns ic1
                on ic1.object_id = c1.object_id
                and ic1.column_id = c1.column_id
                and ic1.index_column_id = 1
            join sys.columns c2
                on t.object_id = c2.object_id
                and UPPER(c2.name) = 'ID'
            join sys.index_columns ic2
                on ic2.object_id = c2.object_id
                and ic2.column_id = c2.column_id
                and ic2.index_id = ic1.index_id
                and ic2.index_column_id = 2
      where UPPER(t.name) = 'OFFLINE_EXECUTION_JOBS')
BEGIN
  CREATE INDEX IX_OFFLINE_EXEC_JOB_ENDTIME_ID ON OFFLINE_EXECUTION_JOBS(END_TIME, ID);
END;



IF OBJECT_ID ( 'dbo.sp_purge_execution_jobs_delete_chunk', 'P' ) IS NOT NULL 
    DROP PROCEDURE dbo.sp_purge_execution_jobs_delete_chunk;
GO

CREATE PROCEDURE dbo.sp_purge_execution_jobs_delete_chunk
  @sql_part nvarchar(max),
  @delete_chunk_size int,
  @time_delay int = null,
  @p1 datetime
AS
DECLARE 
  @rows_deleted int,
  @sql nvarchar(max),
  @delay_stmt nvarchar(300) = N'';
BEGIN
  SET NOCOUNT ON;
  SET IMPLICIT_TRANSACTIONS OFF;
  SET XACT_ABORT ON;
  
  if @time_delay is not null
    set @delay_stmt = + N'WAITFOR DELAY ''' 
                        + RIGHT(N'0' + cast(FLOOR(@time_delay/3600.0) as nvarchar), 2) + ':'
                        + RIGHT(N'0' + cast(FLOOR((@time_delay % 3600)/60.0) as nvarchar), 2) + ':'
                        + RIGHT(N'0' + cast(@time_delay % 60 as nvarchar), 2) + ''' ';

  set @sql = @delay_stmt + N' delete top(' + cast(@delete_chunk_size as nvarchar) + ') ' + @sql_part;
  set @rows_deleted = @delete_chunk_size;
  
  WHILE @rows_deleted = @delete_chunk_size BEGIN
    
    EXEC sp_executesql @sql,
            N'@p1 DATETIME',
            @p1;
    
    set @rows_deleted = @@ROWCOUNT;
  END
END
go

IF OBJECT_ID ( 'dbo.sp_purge_execution_jobs', 'P' ) IS NOT NULL 
    DROP PROCEDURE dbo.sp_purge_execution_jobs;
GO

CREATE PROCEDURE dbo.sp_purge_execution_jobs
  @retention_days int = 90,
  @time_limit int = null,
  @time_delay int = null,
  @delete_chunk_size int = 10000
AS
DECLARE 
  @cur_date datetime,
  @cutoff_date datetime,
  @stop_time datetime,
  @sql nvarchar(max);

BEGIN
  SET NOCOUNT ON;
  SET IMPLICIT_TRANSACTIONS OFF;
  SET XACT_ABORT ON;

  set @stop_time = case when @time_limit is null then null else dateadd(second, @time_limit, getdate()) end
  select @cur_date = dateadd(day, 1, cast(min(end_time) as date)) from offline_execution_jobs where end_time is not null;
  set @cutoff_date = dateadd(day, - @retention_days, getdate());

  WHILE @cur_date < @cutoff_date AND (@stop_time is null OR getdate() < @stop_time) BEGIN

    print convert(varchar(30), getdate(), 109) + ': deleting offline execution jobs older than ' + cast(@cur_date as varchar(30)) + ' ...';

    exec sp_purge_execution_jobs_delete_chunk N'from offline_distribution_events where id in (select e.id from offline_execution_jobs j, offline_flow_events e where j.id = e.job_id and j.end_time < @p1) OPTION(loop join)', @delete_chunk_size, @time_delay, @cur_date;

    exec sp_purge_execution_jobs_delete_chunk N'from offline_exec_param_links where job_id in (select id from offline_execution_jobs where end_time < @p1) OPTION(loop join)', @delete_chunk_size, @time_delay, @cur_date;
    exec sp_purge_execution_jobs_delete_chunk N'from offline_exec_servers where job_id in (select id from offline_execution_jobs where end_time < @p1) OPTION(loop join)', @delete_chunk_size, @time_delay, @cur_date;
    exec sp_purge_execution_jobs_delete_chunk N'from offline_flow_events where job_id in (select id from offline_execution_jobs where end_time < @p1) OPTION(loop join)', @delete_chunk_size, @time_delay, @cur_date;
    exec sp_purge_execution_jobs_delete_chunk N'from offline_manual_operations where job_id in (select id from offline_execution_jobs where end_time < @p1) OPTION(loop join)', @delete_chunk_size, @time_delay, @cur_date;
    exec sp_purge_execution_jobs_delete_chunk N'from offline_parameter_requests where job_id in (select id from offline_execution_jobs where end_time < @p1) OPTION(loop join)', @delete_chunk_size, @time_delay, @cur_date;
    exec sp_purge_execution_jobs_delete_chunk N'from offline_parameters where job_id in (select id from offline_execution_jobs where end_time < @p1) OPTION(loop join)', @delete_chunk_size, @time_delay, @cur_date;
    exec sp_purge_execution_jobs_delete_chunk N'from offline_propagation_events where job_id in (select id from offline_execution_jobs where end_time < @p1) OPTION(loop join)', @delete_chunk_size, @time_delay, @cur_date;
    exec sp_purge_execution_jobs_delete_chunk N'from offline_step_events where job_id in (select id from offline_execution_jobs where end_time < @p1) OPTION(loop join)', @delete_chunk_size, @time_delay, @cur_date;
    exec sp_purge_execution_jobs_delete_chunk N'from offline_steps where job_id in (select id from offline_execution_jobs where end_time < @p1) OPTION(loop join)', @delete_chunk_size, @time_delay, @cur_date;
    exec sp_purge_execution_jobs_delete_chunk N'from offline_user_events where job_id in (select id from offline_execution_jobs where end_time < @p1) OPTION(loop join)', @delete_chunk_size, @time_delay, @cur_date;

    exec sp_purge_execution_jobs_delete_chunk N'from offline_execution_jobs where end_time < @p1', @delete_chunk_size, @time_delay, @cur_date;

    set @cur_date = dateadd(day, 1, @cur_date);
  END;

  if (getdate() > @stop_time)
    print 'Exceeded Time limit.  Exiting procedure.'
  else
    print 'Purge done.'
END;
GO

